import pytest
from slash.utils.marks import get_marks, mark, try_get_mark


def test_mark(marked_obj, mark_name, mark_value):
    assert get_marks(marked_obj) == {mark_name: mark_value}


def test_mark_immutable(marked_obj, mark_name, mark_value):
    marks = get_marks(marked_obj)
    marks["1"] = ["2"]
    assert get_marks(marked_obj) == {mark_name: mark_value}


def test_try_get_mark(marked_obj, mark_name, mark_value):
    assert try_get_mark(marked_obj, mark_name) == mark_value


def test_try_get_mark_fail(marked_obj):
    assert try_get_mark(marked_obj, "nonexistent") is None


def test_try_get_mark_fail_with_default(marked_obj):
    _default = object()
    assert try_get_mark(marked_obj, "nonexistent", _default) is _default


@pytest.mark.parametrize('obj', [1, None, object(), type, "string"])
def test_try_get_mark_fail_non_marked(obj):
    assert try_get_mark(obj, "mark") is None

### Boilerplate ###

mark_factories = []

markfactory = mark_factories.append


@markfactory
def markfactory_class(mark_name, mark_value):
    @mark(mark_name, mark_value)
    class Blap(object):
        pass

    return Blap


@markfactory
def markfactory_func(mark_name, mark_value):
    @mark(mark_name, mark_value)
    def func():
        pass

    return func


@markfactory
def method(mark_name, mark_value):
    class Blap(object):

        @mark(mark_name, mark_value)
        def func(self):
            pass

    return Blap.func


@pytest.fixture(params=mark_factories)
def marked_obj(request, mark_name, mark_value):
    returned = request.param(mark_name, mark_value)
    return returned


@pytest.fixture(params=['mark_name'])
def mark_name(request):
    return request.param


@pytest.fixture(params=['mark_value', 1, True, 1.0])
def mark_value(request):
    return request.param



import itertools

import pytest
from slash.utils.iteration import iteration, PeekableIterator, iter_cartesian_dicts
from slash._compat import iteritems


def test_iteration(objects):
    for index, i in enumerate(iteration(objects)):
        assert i.last_counter0 + 1 == i.last_counter1 == i.total
        assert i.counter0 == index
        assert i.counter1 == index + 1
        if index == 0:
            assert i.first
        else:
            assert not i.first
        assert i.element is objects[index]
        assert index <= len(objects) - 1
        if index == len(objects) - 1:
            assert i.last
        else:
            assert not i.last


def test_iteration_unpacking(objects):
    for index, (i, obj) in enumerate(iteration(objects)):
        assert index == i.counter0
        assert obj is i.element
        assert obj is objects[i.counter0]


def test_iteration_unsupported_sizing():
    for i in iteration(x for x in itertools.count()):
        assert i.first
        with pytest.raises(NotImplementedError):
            i.last
        assert i.last_counter0 is None
        assert i.last_counter1 is None
        break


def test_peekable_iterator(objects):
    it = PeekableIterator(objects)
    for i, x in enumerate(it):
        assert x is objects[i]
        # no matter how many times we peek, we get the same result
        for peek_num in range(3):
            if i == len(objects) - 1:
                assert not it.has_next()
                with pytest.raises(StopIteration):
                    it.peek()
                assert it.peek_or_none() is None
            else:
                assert (it.has_next())
                assert it.peek() is objects[(i + 1)]
                assert it.peek_or_none() is objects[(i + 1)]


def test_cartesian_dict():

    params = {
        'a': [1, 2, 3],
        'b': [4, 5],
        'c': [6, 7],
    }

    assert set(frozenset(iteritems(x)) for x in iter_cartesian_dicts(params)) == \
        set(frozenset([('a', a_value), ('b', b_value), ('c', c_value)])
            for a_value, b_value, c_value in itertools.product(params['a'], params['b'], params['c']))


@pytest.fixture(params=[True, False])
def use_iterator(request):
    return request.param


@pytest.fixture
def objects(use_iterator):
    if use_iterator:
        return range(10)
    return [object() for i in range(10)]



import functools
import itertools
import random

import pytest
import slash
from slash._compat import StringIO
from slash.utils.iteration import iter_cartesian_dicts
from .utils.code_formatter import CodeFormatter


def test_representation_parameters(fixture_store, parametrized_func, params, param_names):

    variations = list(
        fixture_store.iter_parametrization_variations(funcs=[parametrized_func]))

    cartesian_product = list(iter_cartesian_dicts(params))

    assert len(variations) == len(cartesian_product)
    variation_names = set(v.representation for v in variations)
    assert len(variation_names) == len(cartesian_product)
    assert variation_names == set(', '.join('{0}={1}'.format(name, combination[name]) for name in param_names)
                                  for combination in cartesian_product)

def test_representation_fixtures(fixture_store):

    first_param_values = [1, 2, 3]
    second_param_values = [4, 5, 6]

    @fixture_store.add_fixture
    @slash.fixture
    @slash.parametrize('value', first_param_values)
    def first_fixture(value):
        return value

    @fixture_store.add_fixture
    @slash.fixture
    @slash.parametrize('value', second_param_values)
    def second_fixture(value):
        return value

    def test_func(first_fixture, second_fixture):
        pass

    fixture_store.resolve()

    variations = list(
        fixture_store.iter_parametrization_variations(funcs=[test_func]))

    assert len(variations) == len(first_param_values) * len(second_param_values)

    assert set(variation.representation for variation in variations) == set(
        'first_fixture=first_fixture{0}, second_fixture=second_fixture{1}'.format(i, j)
        for i, j in itertools.product(range(len(first_param_values)), range(len(second_param_values))))

@pytest.fixture
def parametrized_func(params, param_names):
    buff = StringIO()
    formatter = CodeFormatter(buff)
    formatter.writeln('def f({0}):'.format(', '.join(param_names)))
    with formatter.indented():
        formatter.writeln('pass')
    globs = {}
    exec(buff.getvalue(), globs)
    returned = globs['f']
    for param_name in param_names:
        returned = slash.parametrize(
            param_name, list(params[param_name]))(returned)
    return returned


@pytest.fixture(
    params=[
        sorted,
        functools.partial(sorted, reverse=True),
        functools.partial(sorted, key=lambda x: random.random()),
    ])
def param_names(request, params):
    sorter = request.param
    return sorter(params)


@pytest.fixture
def params():
    return {
        'a': [1, 2],
        'b': [4, 5],
        'c': [6, 7],
        'd': [8, 9],
    }


@pytest.fixture
def fixture_store():
    return slash.core.fixtures.fixture_store.FixtureStore()


@pytest.fixture(params=[str, repr])
def stringify(request):
    return request.param



import argparse
import os
import sys
from contextlib import contextmanager

from .. import conf, plugins
from .._compat import cStringIO, iteritems, itervalues
from .formatter import Formatter


@contextmanager
def get_cli_environment_context(argv=None, config=conf.config, extra_args=(), positionals_metavar=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = SlashArgumentParser(prog=_deduce_program_name(), positionals_metavar=positionals_metavar)
    if extra_args:
        _populate_extra_args(parser, extra_args)
    argv = list(argv) # copy the arguments, as we'll be gradually removing known arguments
    with _get_active_plugins_context(argv):
        _configure_parser_by_active_plugins(parser)
        _configure_parser_by_config(parser, config)
        if positionals_metavar is not None:
            parsed_args, positionals = parser.parse_known_args(argv)
        else:
            parsed_args = parser.parse_args(argv)
            positionals = []
        parsed_args.positionals = positionals
        _configure_plugins_from_args(parsed_args)
        with _get_modified_configuration_from_args_context(parser, config, parsed_args):
            yield parser, parsed_args

def _deduce_program_name():
    returned = os.path.basename(sys.argv[0])
    if len(sys.argv) > 1:
        returned += " {0}".format(sys.argv[1])
    return returned

def _populate_extra_args(parser, extra_args):
    for argument in extra_args:
        parser.add_argument(*argument.args, **argument.kwargs)

@contextmanager
def _get_active_plugins_context(argv):
    cleanups = []
    prev_active = set(plugins.manager.get_active_plugins())
    try:
        new_active, new_argv = _get_new_active_plugins_from_args(argv)
        for plugin_name in new_active - prev_active:
            plugins.manager.activate(plugin_name)
            cleanups.append((plugins.manager.deactivate, plugin_name))
        for plugin_name in prev_active - new_active:
            plugins.manager.deactivate(plugin_name)
            cleanups.append((plugins.manager.activate, plugin_name))
        del argv[:]
        argv.extend(new_argv)
        yield
    finally:
        for cleanup_func, plugin_name in reversed(cleanups):
            cleanup_func(plugin_name)

_PLUGIN_ACTIVATION_PREFIX = "--with-"
_PLUGIN_DEACTIVATION_PREFIX = "--without-"
def _get_new_active_plugins_from_args(argv):
    new_active = set(plugins.manager.get_active_plugins())
    returned_argv = []
    for arg in argv:
        if arg.startswith(_PLUGIN_DEACTIVATION_PREFIX):
            new_active.discard(arg[len(_PLUGIN_DEACTIVATION_PREFIX):])
        elif arg.startswith(_PLUGIN_ACTIVATION_PREFIX):
            new_active.add(arg[len(_PLUGIN_ACTIVATION_PREFIX):])
        else:
            returned_argv.append(arg)
    return new_active, returned_argv

def _configure_parser_by_active_plugins(parser):
    for plugin in itervalues(plugins.manager.get_active_plugins()):
        plugin.configure_argument_parser(parser)

def _configure_plugins_from_args(args):
    for plugin in itervalues(plugins.manager.get_active_plugins()):
        plugin.configure_from_parsed_args(args)

def _configure_parser_by_config(parser, config):
    parser.add_argument(
        "-o", dest="config_overrides", metavar="PATH=VALUE", action="append",
        default=[],
        help="Provide overrides for configuration"
    )
    for path, node, cmdline in _iter_cmdline_config(config):
        cmdline.configure_parser(parser, path, node)

def _iter_cmdline_config(config):
    for path, cfg in config.traverse_leaves():
        cmdline = (cfg.metadata or {}).get("cmdline")
        if cmdline is None:
            continue
        yield path, cfg, cmdline

@contextmanager
def _get_modified_configuration_from_args_context(parser, config, args):
    to_restore = []
    try:
        for path, cfg, cmdline in _iter_cmdline_config(config):
            old_value = cfg.get_value()
            new_value = cmdline.update_value(old_value, args)
            if new_value != old_value:
                to_restore.append((path, cfg.get_value()))
                config.assign_path(path, new_value, deduce_type=True, default_type=str)
        for override in args.config_overrides:
            if "=" not in override:
                parser.error("Invalid config override: {0}".format(override))
            path, _ = override.split("=", 1)
            to_restore.append((path, config.get_path(path)))
            try:
                config.assign_path_expression(override, deduce_type=True, default_type=str)
            except ValueError:
                parser.error("Invalid value for config override: {0}".format(override))
        yield
    finally:
        for path, prev_value in reversed(to_restore):
            config.assign_path(path, prev_value)

class SlashArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        positionals_metavar = kwargs.pop("positionals_metavar", None)
        super(SlashArgumentParser, self).__init__(*args, **kwargs)
        self._positionals_metavar = positionals_metavar

    def format_help(self):
        returned = cStringIO()
        helpstring = super(SlashArgumentParser, self).format_help()
        helpstring = self._tweak_usage_positional_metavars(helpstring)
        returned.write(helpstring)
        f = Formatter(returned)
        for index, (plugin_name, plugin) in enumerate(self._iter_available_plugins()):
            if index == 0:
                f.writeln()
                f.writeln("Available (inactive) plugins:")
                f.indent()
            f.write(_PLUGIN_ACTIVATION_PREFIX + plugin_name)
            description = plugin.get_description()
            if description is not None:
                f.writeln()
                with f.indented(2):
                    f.write(description)
            f.writeln()
        return returned.getvalue()

    def _tweak_usage_positional_metavars(self, usage):
        """
        Adds fake positionals metavar at the end of the usage line
        """
        if self._positionals_metavar is None:
            return usage

        returned = ""
        added_metavars = False
        for line in cStringIO(usage):
            if not added_metavars and not line.strip():
                returned = returned[:-1] + " [{0} [{0} ...]] \n".format(self._positionals_metavar)
            returned += line
        return returned

    def _iter_available_plugins(self):
        active_plugin_names = set(plugins.manager.get_active_plugins())
        for plugin_name, plugin in iteritems(plugins.manager.get_installed_plugins()):
            if plugin_name not in active_plugin_names:
                yield plugin_name, plugin

class Argument(object):
    """
    helper to defer initialization of cmdline parsers to later stages
    """
    def __init__(self, *args, **kwargs):
        super(Argument, self).__init__()
        self.args = args
        self.kwargs = kwargs



import itertools
import os
import random
import shutil
import tempfile

from forge import Forge

import gossip
import pytest
import slash
from slash import resuming
from slash.loader import Loader

from .utils.cartesian import Cartesian
from .utils.suite_writer import Suite


@pytest.fixture(scope='session', autouse=True)
def unittest_mode_logging():
    slash.config.root.log.unittest_mode = True


@pytest.fixture(scope='session', autouse=True)
def random_seed():
    random.seed(0xdeadface)


@pytest.fixture(scope='session', autouse=True)
def no_user_config(request):
    tmpdir = tempfile.mkdtemp()
    slash.conf.config.root.run.user_customization_file_path = os.path.join(
        tmpdir, 'slashrc')

    @request.addfinalizer
    def cleanup():
        os.rmdir(tmpdir)


@pytest.fixture
def forge(request):

    returned = Forge()

    @request.addfinalizer
    def cleanup():
        returned.verify()
        returned.restore_all_replacements()

    return returned


@pytest.fixture
def config_override(request):

    def _override(path, value):
        prev_value = slash.config.get_config(path).get_value()

        @request.addfinalizer
        def restore():
            slash.config.assign_path(path, prev_value)
        slash.config.assign_path(path, value)
    return _override


@pytest.fixture
def cartesian():
    return Cartesian()


@pytest.fixture(scope="function", autouse=True)
def cleanup_hook_registrations(request):
    @request.addfinalizer
    def _cleanup():
        for hook in gossip.get_group("slash").get_hooks():
            hook.unregister_all()
        assert not gossip.get_group("slash").get_subgroups()


@pytest.fixture(scope="function")
def checkpoint():
    return Checkpoint()

_timestamp = itertools.count(1000000)


class Checkpoint(object):

    called_count = 0
    args = kwargs = timestamp = None

    def __call__(self, *args, **kwargs):
        self.called_count += 1
        self.args = args
        self.kwargs = kwargs
        self.timestamp = next(_timestamp)

    @property
    def called(self):
        return self.called_count > 0


@pytest.fixture(autouse=True, scope="function")
def fix_resume_path(request):
    prev = resuming._RESUME_DIR
    resuming._RESUME_DIR = tempfile.mkdtemp()

    @request.addfinalizer
    def cleanup():
        shutil.rmtree(resuming._RESUME_DIR)
        resuming._RESUME_DIR = prev


@pytest.fixture
def suite_test(suite, test_type, is_last_test):
    returned = suite.add_test(type=test_type)
    if not is_last_test:
        _ = suite.add_test(type=test_type)

    return returned


@pytest.fixture(params=[True, False])
def is_last_test(request):
    return request.param


@pytest.fixture(params=['method', 'function'])
def test_type(request):
    return request.param


@pytest.fixture
def suite():
    returned = Suite()
    returned.populate()
    return returned


@pytest.fixture
def slash_session():
    return slash.Session()


@pytest.fixture
def test_loader():
    return Loader()


@pytest.fixture
def active_slash_session(request):
    returned = slash.Session()
    returned.__enter__()

    @request.addfinalizer
    def finalize():
        returned.__exit__(None, None, None)

    return returned



import sys
from contextlib import contextmanager

from . import hooks as trigger_hook
from . import site
from .core.session import Session
from .conf import config
from .loader import Loader
from .reporting.console_reporter import ConsoleReporter
from .utils import cli_utils


class Application(object):
    def __init__(self, parser, args, report_stream):
        super(Application, self).__init__()
        self.parser = parser
        self.args = args
        self.report_stream = report_stream
        self.test_loader = Loader()
        self.session = Session(reporter=ConsoleReporter(level=config.root.log.console_level, stream=report_stream), console_stream=report_stream)

    def error(self, message, usage=True):
        if usage:
            self.parser.error(message)
        else:
            sys.exit('Error: {0}'.format(message))

@contextmanager
def get_application_context(parser=None, argv=None, args=(), report_stream=sys.stderr, enable_interactive=False, positionals_metavar=None):
    site.load()
    args = list(args)
    if enable_interactive:
        args.append(
            cli_utils.Argument("-i", "--interactive", help="Enter an interactive shell",
                               action="store_true", default=False)
        )
    with cli_utils.get_cli_environment_context(argv=argv, extra_args=args, positionals_metavar=positionals_metavar) as (parser, parsed_args):
        app = Application(parser=parser, args=parsed_args, report_stream=report_stream)
        _check_unknown_switches(app)
        with app.session:
            yield app
            trigger_hook.result_summary()  # pylint: disable=no-member

def _check_unknown_switches(app):
    unknown = [arg for arg in app.args.positionals if arg.startswith("-")]
    if unknown:
        app.error("Unknown flags: {0}".format(", ".join(unknown)))



import os

import gossip
import pytest
import slash
from slash._compat import PY2
from slash import hooks, plugins
from slash.plugins import IncompatiblePlugin, PluginInterface

from .utils import CustomException, TestCase


@pytest.fixture
def restore_plugins_on_cleanup(request):
    request.addfinalizer(plugins.manager.install_builtin_plugins)
    request.addfinalizer(plugins.manager.uninstall_all)


@pytest.fixture(autouse=True, scope="function")
def reset_gossip(request):
    @request.addfinalizer
    def cleanup():
        for group in list(gossip.get_groups()):
            if group.name == 'slash':
                continue
            group.undefine()

        for hook in gossip.get_all_hooks():
            if hook.group.name != 'slash':
                hook.undefine()
            else:
                hook.unregister_all()


def test_registers_on_none(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on(None)
        def some_method_here(self):
            checkpoint()

    gossip.trigger('slash.some_method_here')
    assert not checkpoint.called



def test_registers_on_with_private_methods(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on('some_hook')
        def _handler(self):
            checkpoint()

    assert not checkpoint.called
    gossip.trigger('some_hook')
    assert checkpoint.called


def test_class_variables_allowed(restore_plugins_on_cleanup):
    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        ATTRIBUTE = 'some_value'

        def get_name(self):
            return 'sample'


def test_active_decorator(restore_plugins_on_cleanup):

    plugins.manager.uninstall_all()

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

    assert isinstance(SamplePlugin, type)
    assert issubclass(SamplePlugin, PluginInterface)
    [active] = plugins.manager.get_active_plugins().values()
    assert isinstance(active, SamplePlugin)


def test_custom_hook_registration():

    hook_name = 'some_hook'
    with pytest.raises(LookupError):
        gossip.get_hook(hook_name)

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass
    p = MyPlugin()
    plugins.manager.install(p, activate=True)
    registrations = gossip.get_hook(hook_name).get_registrations()
    assert 1 == len(registrations)
    [r] = registrations
    if PY2:
        assert r.func.__func__ is MyPlugin.unknown.__func__
    else:
        assert r.func.__func__ is MyPlugin.unknown

    # make sure we deactivate properly as well
    plugins.manager.deactivate(p)
    assert not gossip.get_hook(hook_name).get_registrations()


def test_register_invalid_hook():

    initially_installed = list(plugins.manager.get_installed_plugins())

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert list(plugins.manager.get_installed_plugins()) == initially_installed


def test_register_custom_hooks_strict_group():

    initially_installed = list(plugins.manager.get_installed_plugins())

    hook_name = "some_group.some_hook"
    gossip.get_or_create_group("some_group").set_strict()

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert list(plugins.manager.get_installed_plugins()) == initially_installed


class BuiltinPluginsTest(TestCase):

    def test_hooks_start_condition(self):
        "make sure that all hooks are either empty, or contain callbacks marked with `slash.<identifier>`"
        for hook_name, hook in hooks.get_all_hooks():
            for registration in hook.get_registrations():
                self.assertTrue(
                    registration.token.startswith("slash."),
                    "Callback {0}.{1} is not a builtin!".format(hook_name, identifier)
                )

    def test_builtin_plugins_are_installed(self):
        installed = plugins.manager.get_installed_plugins()
        self.assertNotEquals(installed, {})
        for filename in os.listdir(os.path.join(os.path.dirname(plugins.__file__), "builtin")):
            if filename.startswith("_") or filename.startswith(".") or not filename.endswith(".py"):
                continue
            self.assertIn(filename[:-3], installed)


class PluginInstallationTest(TestCase):

    def test_cannot_install_incompatible_subclasses(self):
        plugins.manager.uninstall_all()
        self.addCleanup(plugins.manager.install_builtin_plugins)

        class Incompatible(object):
            pass
        for invalid in (Incompatible, Incompatible(), PluginInterface, object(), 1, "string"):
            with self.assertRaises(IncompatiblePlugin):
                plugins.manager.install(invalid)
        self.assertEquals(plugins.manager.get_installed_plugins(), {})

    def test_install_uninstall(self):
        plugin_name = "some_plugin_name"

        class CustomPlugin(PluginInterface):

            def get_name(self):
                return plugin_name
        with self.assertRaises(LookupError):
            plugins.manager.get_plugin(plugin_name)
        plugin = CustomPlugin()
        plugins.manager.install(plugin)
        self.assertIs(plugins.manager.get_plugin(plugin_name), plugin)
        plugins.manager.uninstall(plugin)
        with self.assertRaises(LookupError):
            plugins.manager.get_plugin(plugin_name)


class PluginDiscoveryTest(TestCase):

    def setUp(self):
        super(PluginDiscoveryTest, self).setUp()
        self.root_path = self.get_new_path()
        self.expected_names = set()
        for index, path in enumerate([
                "a/b/p1.py",
                "a/b/p2.py",
                "a/p3.py",
                "a/b/c/p4.py",
        ]):
            plugin_name = "auto_plugin_{0}".format(index)
            path = os.path.join(self.root_path, path)
            if not os.path.isdir(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            with open(path, "w") as f:
                f.write("""
import slash.plugins
from slash.plugins.interface import PluginInterface

class {name}(PluginInterface):
    def get_name(self):
        return {name!r}

def install_plugins():
""".format(name=plugin_name))
                if index % 2 == 0:
                    # don't install
                    f.write("     pass")
                else:
                    self.expected_names.add(plugin_name)
                    f.write("     slash.plugins.manager.install({name}())".format(name=plugin_name))
        for junk_file in [
                "a/junk1.p",
                "a/b/junk2",
                "a/b/c/junk3",
        ]:
            with open(os.path.join(self.root_path, junk_file), "w") as f:
                f.write("---JUNK----")
        self.override_config("plugins.search_paths", [self.root_path])

    def tearDown(self):
        plugins.manager.uninstall_all()
        plugins.manager.install_builtin_plugins()
        super(PluginDiscoveryTest, self).tearDown()

    def test_discovery(self):
        plugins.manager.uninstall_all()
        self.addCleanup(plugins.manager.install_builtin_plugins)
        plugins.manager.discover()
        self.assertEquals(
            set(plugins.manager.get_installed_plugins().keys()),
            self.expected_names
        )


class PluginActivationTest(TestCase):

    def setUp(self):
        super(PluginActivationTest, self).setUp()
        self.plugin = StartSessionPlugin()

    def test_get_active_plugins(self):
        plugins.manager.install(self.plugin)
        self.addCleanup(plugins.manager.uninstall, self.plugin)
        self.assertEquals(plugins.manager.get_active_plugins(), {})
        plugins.manager.activate(self.plugin)
        self.assertEquals(
            plugins.manager.get_active_plugins(),
            {self.plugin.get_name(): self.plugin}
        )
        plugins.manager.deactivate(self.plugin)
        self.assertEquals(plugins.manager.get_active_plugins(), {})

    def test_deactivaion_no_activation(self):
        plugins.manager.install(self.plugin)
        self.assertFalse(self.plugin._deactivate_called)
        plugins.manager.uninstall(self.plugin)
        self.assertFalse(self.plugin._deactivate_called, "Deactivate called even though plugin not activated")

    def test_activation_exception(self):
        self.plugin.activate = CustomException.do_raise
        plugins.manager.install(self.plugin)
        self.addCleanup(plugins.manager.uninstall, self.plugin)

        with self.assertRaises(CustomException):
            plugins.manager.activate(self.plugin)

        # make sure no registrations are in effect...
        self.assert_hooks_not_registered()

        plugins.manager.deactivate(self.plugin)
        self.assertFalse(self.plugin._deactivate_called, "Deactivate unexpectedly called!")

    def test_deactivation_exception(self):
        self.plugin.deactivate = CustomException.do_raise
        plugins.manager.install(self.plugin, activate=True)
        self.addCleanup(plugins.manager.uninstall, self.plugin)

        with self.assertRaises(CustomException):
            plugins.manager.deactivate(self.plugin)
        self.assert_hooks_not_registered()

    def test_activate_called(self):
        plugins.manager.install(self.plugin)
        self.addCleanup(plugins.manager.uninstall, self.plugin)
        self.assertFalse(self.plugin._activate_called)
        plugins.manager.activate(self.plugin)
        self.assertTrue(self.plugin._activate_called)

    def test_deactivate_called_on_deactivate(self):
        plugins.manager.install(self.plugin)
        self.assertFalse(self.plugin._deactivate_called)
        plugins.manager.activate(self.plugin)
        self.assertFalse(self.plugin._deactivate_called)
        plugins.manager.deactivate(self.plugin)
        self.assertTrue(self.plugin._deactivate_called)

    def test_hook_registration(self):
        plugins.manager.install(self.plugin)
        self.addCleanup(plugins.manager.uninstall, self.plugin)
        self.assert_hooks_not_registered()
        plugins.manager.activate(self.plugin)
        hooks.session_start()
        self.assertEquals(self.plugin.session_start_call_count, 1)
        plugins.manager.deactivate(self.plugin)
        hooks.session_start()
        self.assertEquals(self.plugin.session_start_call_count, 1)

    def test_install_and_activate(self):
        """test plugins.manager.install(..., activate=True)"""
        plugins.manager.install(self.plugin, activate=True)
        self.addCleanup(plugins.manager.uninstall, self.plugin)
        self.assertIn(self.plugin.get_name(), plugins.manager.get_active_plugins())

    def test_uninstall_also_deactivates(self):
        plugins.manager.install(self.plugin)
        plugins.manager.activate(self.plugin)
        plugins.manager.uninstall(self.plugin)
        hooks.session_start()
        self.assertEquals(self.plugin.session_start_call_count, 0)

    def test_cannot_activate_uninstalled_plugin(self):
        class Plugin(PluginInterface):

            def get_name(self):
                return "Test plugin"
        with self.assertRaisesRegexp(ValueError, ".*not installed.*"):
            plugins.manager.activate(Plugin())

    def test_unknown_hook_names(self):
        "Make sure that plugins with unknown hook names get discarded"
        class Plugin(PluginInterface):

            def get_name(self):
                return "Test plugin"

            def unknown_hook_1(self):
                pass

        plugin = Plugin()
        plugins.manager.install(plugin)
        self.addCleanup(plugins.manager.uninstall, plugin)
        with self.assertRaisesRegexp(IncompatiblePlugin, r"\bUnknown hooks\b.*"):
            plugins.manager.activate(plugin)

    def test_custom_hook_names(self):
        "Make sure that plugins with unknown hook names get discarded"
        class Plugin(PluginInterface):

            def get_name(self):
                return "Test plugin"

            def custom_hook(self):
                pass

        hooks.add_custom_hook("custom_hook")
        self.addCleanup(hooks.remove_custom_hook, "custom_hook")
        plugin = Plugin()
        plugins.manager.install(plugin, activate=True)
        self.addCleanup(plugins.manager.uninstall, plugin)

    def assert_hooks_not_registered(self):
        hooks.session_start()
        self.assertEquals(self.plugin.session_start_call_count, 0, "Hook unexpectedly registered!")


class StartSessionPlugin(PluginInterface):
    _activate_called = False
    _deactivate_called = False

    def __init__(self):
        super(StartSessionPlugin, self).__init__()
        self.session_start_call_count = 0

    def get_name(self):
        return "start-session"

    def session_start(self):
        self.session_start_call_count += 1

    def activate(self):
        self._activate_called = True

    def deactivate(self):
        self._deactivate_called = True



from contextlib import contextmanager

from forge import Anything

import pytest
from slash import exception_handling
from slash.exceptions import SkipTest
from slash.utils import debug

from .utils import CustomException, TestCase


def test_passthrough_types():

    value = CustomException()

    with pytest.raises(CustomException) as caught:
        with exception_handling.handling_exceptions(passthrough_types=(CustomException,)):
            raise value
    assert value is caught.value
    assert not exception_handling.is_exception_handled(value)

    with pytest.raises(CustomException) as caught:
        with exception_handling.handling_exceptions(passthrough_types=(AttributeError,)):
            raise value
    assert value is caught.value
    assert exception_handling.is_exception_handled(value)



class HandlingExceptionsContextTest(TestCase):
    def setUp(self):
        super(HandlingExceptionsContextTest, self).setUp()
        self.handler = self.forge.create_wildcard_mock()
        self.forge.replace_with(exception_handling, "_EXCEPTION_HANDLERS", [self.handler])
        # expect handler to be called once!
        self.raised = CustomException()
        self.handler((CustomException, self.raised, Anything()))
        self.forge.replay()
    def test_handling_exceptions(self):
        with self.assertRaises(CustomException) as caught:
            with exception_handling.handling_exceptions():
                with exception_handling.handling_exceptions():
                    with exception_handling.handling_exceptions():
                        raise self.raised
        self.assertIs(caught.exception, self.raised)

class DebuggingTest(TestCase):
    def setUp(self):
        super(DebuggingTest, self).setUp()
        self.forge.replace_with(debug, "launch_debugger", self.dummy_debugger)
        self.debugger_called = False
    def dummy_debugger(self, *args, **kwargs):
        self.debugger_called = True
    def test_debugging_not_configured(self):
        self._raise_exception_in_context(ZeroDivisionError)
        self.assertFalse(self.debugger_called)
    def test_debugging_configured_no_skips(self):
        self.override_config("debug.debug_skips", False)
        self.override_config("debug.enabled", True)
        self._raise_exception_in_context(SkipTest)
        self.assertFalse(self.debugger_called)
    def test_debugging_skips(self):
        self.override_config("debug.debug_skips", True)
        self.override_config("debug.enabled", True)
        self._raise_exception_in_context(SkipTest)
        self.assertTrue(self.debugger_called)
    def _raise_exception_in_context(self, exception_type):
        with self.assertRaises(exception_type):
            with exception_handling.handling_exceptions():
                raise exception_type()

class ExceptionSwallowingTest(TestCase):
    def test_swallow(self):
        with exception_handling.get_exception_swallowing_context():
            raise CustomException("!!!")
    def test_no_swallow(self):
        with self.assertNoSwallow() as raised:
            raise exception_handling.noswallow(raised)
    def test_disable_exception_swallowing_function(self):
        with self.assertNoSwallow() as raised:
            exception_handling.disable_exception_swallowing(raised)
            raise raised
    def test_disable_exception_swallowing_decorator(self):
        @exception_handling.disable_exception_swallowing
        def func():
            raise raised
        with self.assertNoSwallow() as raised:
            func()
    @contextmanager
    def assertNoSwallow(self):
        raised = CustomException()
        with self.assertRaises(CustomException) as caught:
            with exception_handling.get_exception_swallowing_context():
                yield raised
        self.assertIs(raised, caught.exception)



"""Slash should support running ad-hoc test objects, that were not loaded from a file
"""
import slash
from slash.core.runnable_test import RunnableTest


def test_ad_hoc_test_running(checkpoint):

    class SampleTest(RunnableTest):
        def run(self):
            checkpoint()

        def get_requirements(self):
            return []

    with slash.Session() as session:
        with session.get_started_context():
            slash.runner.run_tests([SampleTest()])
    assert session.results.is_success()



import collections

import logbook

from .ctx import context


class SessionWarnings(object):
    """
    Holds all warnings emitted during the session
    """
    def __init__(self):
        super(SessionWarnings, self).__init__()
        self.records = []

    def __iter__(self):
        "Iterates through stored warnings"
        return iter(self.records)

    def __len__(self):
        return len(self.records)

    def __nonzero__(self):
        return bool(self.records)

    __bool__ = __nonzero__

class WarnHandler(logbook.Handler, logbook.StringFormatterHandlerMixin):
    """
    Like a stream handler but keeps the values in memory.
    This logger provides some ways to store warnings to log again at the end of the session.
    """
    default_format_string = (u'[{record.time:%Y-%m-%d %H:%M}] '
      '{record.level_name}: {record.extra[source]}: {record.message}')
    def __init__(self, session_warnings, format_string=None, filter=None, bubble=True):
        logbook.Handler.__init__(self, logbook.WARNING, filter, bubble)
        logbook.StringFormatterHandlerMixin.__init__(self, format_string)
        #: captures the :class:`LogRecord`\s as instances
        self.records = session_warnings.records

    def should_handle(self, record):
        """Returns `True` if this record is a warning """
        return record.level == self.level

    def emit(self, record):
        self.records.append(Warning(record, self.format(record)))

WarningKey = collections.namedtuple("WarningKey", ("filename", "lineno"))

class Warning(object):

    def __init__(self, record, message):
        super(Warning, self).__init__()
        self.details = record.to_dict()
        self.details['session_id'] = context.session_id
        self.details['test_id'] = context.test_id
        self.key = WarningKey(filename=self.details['filename'], lineno=self.details['lineno'])
        self._repr = message

    def to_dict(self):
        return self.details.copy()

    def __repr__(self):
        return self._repr



import os
import sys

import requests

from .conf import config

def load(thing=None):
    """
    Loads site files (customization files) from various locations.

    Without an argument, loads all default site customization options.

    With a specific argument, load that argument as a customization file (real file or URL).
    """
    if thing is None:
        return _load_defaults()
    _load_filename_or_url(thing)

def _load_defaults():
    _load_slashrc()
    _load_local_slashrc()
    _load_environment()
    _load_entry_points()

def _load_slashrc():
    _load_file_if_exists(os.path.expanduser(config.root.run.user_customization_file_path))

def _load_local_slashrc():
    _load_file_if_exists(os.path.abspath(os.path.expanduser(config.root.run.project_customization_file_path)))

def _load_file_if_exists(path):
    if os.path.isfile(path):
        old_sys_path = sys.path[:]
        sys.path.insert(0, os.path.dirname(path))
        try:
            _load_filename(path)
        finally:
            sys.path[:] = old_sys_path

def _load_environment():
    loaded_url_or_file = os.environ.get("SLASH_SETTINGS")
    if loaded_url_or_file:
        load(loaded_url_or_file)

def _load_entry_points():
    import pkg_resources
    for customize_function_loader in pkg_resources.iter_entry_points("slash.site.customize"): # pylint: disable=no-member
        func = customize_function_loader.load()
        func()

def _load_filename_or_url(filename_or_url):
    if os.path.isfile(filename_or_url):
        _load_filename(filename_or_url)
    else:
        _load_url(filename_or_url)

def _load_filename(filename):
    with open(filename, "r") as f:
        _load_source(f.read(), filename)

def _load_url(url):
    response = requests.get(url)
    response.raise_for_status()
    _load_source(response.content, url)

def _load_source(source, filename):
    code = compile(source, os.path.abspath(filename), 'exec')
    exec(code, {"__file__" : filename}) # pylint: disable=W0122



import json
import os

import emport

import dessert
import pytest
from slash.core.error import Error

from .utils import without_pyc

def test_error_exception_str_repr(error):
    assert "NotImplementedError" in str(error)
    assert "NotImplementedError" in repr(error)

def test_detailed_exception(error):
    assert error.get_detailed_str()
    assert 'NotImplementedError' in error.get_detailed_str()

def test_error_filename(error):
    assert error.filename == without_pyc(os.path.abspath(__file__))

def test_error_func_name(error):
    assert error.func_name == "func_3"

def test_code_string(error):
    assert error.cause.code_line == "    raise NotImplementedError()"
    assert error.cause.code_string == """def func_3():

    local_func_3 = global_func_3
    raise NotImplementedError()\n"""

def test_frame_locals(error):
    assert error.traceback.frames[-3].locals == {
        "local_func_1": {
            "value": "'global_func_1'"
        }}

def test_to_list(error):
    serialized = error.traceback.to_list()
    assert serialized[-3]['locals'] == {
        "local_func_1": {
            "value": "'global_func_1'"
        }}
    # Just make sure that it's serializable
    json.dumps(serialized)

def test_frame_locals_no_assertion_markers(assertion_error):
    for var_name, var in assertion_error.cause.locals.items():
        assert "@" not in var_name

def test_frame_globals(error):
    assert error.traceback.frames[-3].globals == {
        "global_func_1": {
            "value": "'global_func_1'"
        }}


def test_capture_exception_twice_caches_object():
    try:
        try:
            raise RuntimeError()
        except RuntimeError:
            err1 = Error.capture_exception()
            raise
    except RuntimeError:
        err2 = Error.capture_exception()

    assert err1 is err2


def test_detailed_traceback(error):
    detailed = error.get_detailed_str()
    assert detailed

####

@pytest.fixture
def error():
    try:
        func_1()
    except:
        return Error.capture_exception()
    else:
        assert False, "Did not fail"

@pytest.fixture
def non_exception_error():
    def func1():
        return func2()
    def func2():
        return Error('some_error')

    err = func1()
    return err


global_func_1 = "global_func_1"
global_func_2 = "global_func_2"
global_func_3 = "global_func_3"

def func_1():
    local_func_1 = global_func_1

    func_2()

def func_2():
    local_func_2 = global_func_2

    func_3()

def func_3():

    local_func_3 = global_func_3
    raise NotImplementedError()

@pytest.fixture
def assertion_error(tmpdir):
    filename = tmpdir.join("file.py")
    filename.write("""
def f(x):
    return x

def g(x):
    return x

def func():
    assert f(g(1)) == g(f(2))""")

    with dessert.rewrite_assertions_context():
        module = emport.import_file(str(filename))

    try:
        module.func()
    except:
        return Error.capture_exception()
    else:
        assert False, "Did not fail"



from ..interface import PluginInterface
from ...conf import config
from ...ctx import session
from functools import partial
import requests


def _post_notification(url, api_key, title, message):
    response = requests.post(url, {"apikey": api_key, "application": "slash", "event": title, "description": message})
    response.raise_for_status()


def _pushbullet_notification(api_key, title, message):
    data = {
        "type": "note",
        "title": title,
        "body": message
    }
    response = requests.post(
        "https://api.pushbullet.com/api/pushes",
        data=data,
        auth=(api_key, ""))
    response.raise_for_status()


class Plugin(PluginInterface):
    def get_name(self):
        return "notifications"

    def activate(self):
        config.extend({
            "plugins": {
                "notifications" : {
                    "prowl_api_key" : None,
                    "nma_api_key" : None,
                    "pushbullet_api_key": None,
            }}})

    def deactivate(self):
        config["plugins"].pop("notifications")

    def session_end(self):
        self._notify("Session Ended", "Session {0} ended".format(session.id))

    def _notify(self, title, message):
        for func, api_key in [
                (partial(_post_notification, "https://prowl.weks.net/publicapi/add"), config.root.plugins.notifications.prowl_api_key),
                (partial(_post_notification, "https://www.notifymyandroid.com/publicapi/notify"), config.root.plugins.notifications.nma_api_key),
                (_pushbullet_notification, config.root.plugins.notifications.pushbullet_api_key),
                ]:
            if api_key is None:
                continue

            func(api_key, title, message)



import pytest
from slash.core.cleanup_manager import CleanupManager

from .conftest import Checkpoint


def test_cleanup_default_scope(cleanup_manager, cleanup):
    cleanup_manager.add_cleanup(cleanup)
    assert not cleanup.called
    cleanup_manager.call_cleanups()
    assert cleanup.called

def test_default_scope(cleanup_manager):
    cleanup1 = cleanup_manager.add_cleanup(Checkpoint())

    with cleanup_manager.scope('session'):

        cleanup2 = cleanup_manager.add_cleanup(Checkpoint())

        with cleanup_manager.scope('module'):

            cleanup3 = cleanup_manager.add_cleanup(Checkpoint())

            assert not cleanup3.called
        assert not cleanup2.called

    assert not cleanup1.called
    cleanup_manager.call_cleanups()
    assert cleanup1.called
    assert cleanup2.called
    assert cleanup3.called
    assert cleanup3.timestamp < cleanup2.timestamp < cleanup1.timestamp


@pytest.fixture
def cleanup_manager():
    returned = CleanupManager()
    returned.push_scope(None)
    return returned


@pytest.fixture
def cleanup(checkpoint):
    return checkpoint



import pytest
import slash

from .utils import run_tests_assert_success, without_pyc


@pytest.mark.parametrize('reprify', [repr, str])
def test_str_repr(warning, reprify):
    assert 'this is a warning' in reprify(warning)

def test_location(warning):
    assert warning.details['filename'] == without_pyc(__file__)

def test_to_dict(warning):
    assert isinstance(warning.to_dict(), dict)

@pytest.fixture
def warning():
    class SampleTest(slash.Test):
        def test(self):
            slash.logger.warning("this is a warning. Param: {0}", 1)

    session = run_tests_assert_success(SampleTest)

    assert len(session.warnings) == 1

    [warning] = session.warnings
    return warning



class EventRecorder(object):
    def __init__(self):
        super(EventRecorder, self).__init__()
        self.events = {}
        self.timestamp = 0
    def record(self, event_name, **kwargs):
        assert event_name not in self.events, "Event {0} already recorded".format(event_name)
        self.timestamp += 1
        self.events[event_name] = Event(self.timestamp, kwargs)
    def __getitem__(self, event_name):
        return self.events[event_name]

class Event(object):
    happened = True
    def __init__(self, timestamp, info):
        super(Event, self).__init__()
        self.timestamp = timestamp
        self.info = info





import itertools


class BalancedStrategy(object):

    def __init__(self, max_tests_per_file=5, max_tests_per_class=3):
        super(BalancedStrategy, self).__init__()
        self._max_tests_per_file = max_tests_per_file
        self._max_tests_per_class = max_tests_per_class
        self._type_gen = itertools.cycle(['function', 'method'])

    def get_test_type(self):
        return next(self._type_gen)

    def get_file_for_test(self, suite):
        candidate = suite.get_last_file()
        if candidate is not None and candidate.get_num_tests() < self._max_tests_per_file:
            return candidate
        return suite.add_file()

    def get_class_for_test(self, file):
        candidate = file.get_last_class()
        if candidate is not None and candidate.get_num_tests() < self._max_tests_per_class:
            return candidate
        return file.add_class()



import pytest


def test_autouse_fixtures_global(tracked_suite):

    fixture = tracked_suite.slashconf.add_fixture(autouse=True)

    results = tracked_suite.run().session.results
    assert len(results) > 0
    for result in results:
        assert fixture.id in result.data['active_fixtures']

def test_autouse_fixtures_specific_module(tracked_suite, suite_test):

    fixture = suite_test.file.add_fixture(autouse=True)
    summary = tracked_suite.run()
    assert len(summary.session.results) > 0
    assert len(tracked_suite.files) > 1

    for test in tracked_suite:
        for result in summary.get_all_results_for_test(test):
            if test.file is suite_test.file:
                assert fixture.id in result.data['active_fixtures']
            else:
                assert fixture.id not in result.data['active_fixtures']


@pytest.fixture
def tracked_suite(suite, suite_test):
    assert len(suite)
    for test in suite:
        test.prepend_line('slash.context.result.data["active_fixtures"] = __ut__.get_fixture_memento()')
    return suite



from .utils import (
    TestCase,
    CustomException,
    )
from contextlib import contextmanager

import pytest

import slash
from slash import should
from slash.exceptions import TestFailed

@pytest.mark.parametrize('pair', [
    (1, 1),
    (1, 1.00000001),
    ])
def test_assert_almost_equal_positive(pair):
    a, b = pair
    slash.assert_almost_equal(a, b)

@pytest.mark.parametrize('combination', [
    (1, 1, 0),
    (1, 1, 0.1),
    (1, 1.1, 0.5),
    (1.0001, 1.00009, 0.00002),
    ])
def test_assert_almost_equal_positive_with_delta(combination):
    a, b, delta = combination
    slash.assert_almost_equal(a, b, delta)

@pytest.mark.parametrize('combination', [
    (1, 12, 0),
    (1, 1.1, 0.00001),
    (1.0001, 1.00009, 0.000001),
    ])
def test_assert_almost_equal_negative_with_delta(combination):
    a, b, delta = combination
    with pytest.raises(AssertionError):
        slash.assert_almost_equal(a, b, delta)



class AssertionsTest(TestCase):

    def test_assert_equals(self):
        with checking(should.equal, should.not_equal):
            good(1, 1)
            bad(1, 2)

    def test_assert_empty(self):
        with checking(should.be_empty, should.not_be_empty):
            good([])
            good({})
            good(set())
            bad([1, 2, 3])
            bad({1: 2})
            bad(set([1, 2, 3]))

    def test_assert_isinstance(self):
        with checking(should.be_a, should.not_be_a):
            good(1, int)
            good("a", str)
            good({}, dict)
            bad(1, 1)
            bad(1, str)
            bad(None, str)
    def test_is_none(self):
        with checking(should.be_none, should.not_be_none):
            good(None)
            bad("None")
    def test_is(self):
        obj = object()
        with checking(should.be, should.not_be):
            good(obj, obj)
            good(self, self)
            bad(obj, object())
            bad({}, {})
    def test_truth(self):
        with checking(should.be_true, should.be_false):
            good(True)
            good("hello")
            bad(False)
            bad("")
            bad({})
    def test_in(self):
        with checking(should.be_in, should.not_be_in):
            good(1, [1, 2, 3])
            good("e", "hello")
            bad(1, [])
            bad("e", "fdfd")
        with checking(should.contain, should.not_contain):
            good([1, 2, 3], 1)
            good("hello", "e")
            bad("fdfdfd", "e")
    def test_raises_exception(self):
        thrown = CustomException()
        with should.raise_exception(CustomException) as caught:
            raise thrown
        self.assertIs(caught.exception, thrown)
        try:
            with should.raise_exception(CustomException):
                raise OtherException()
        except OtherException:
            pass
        else:
            self.fail("should.raise_exception allowed a different type of exception to be raised")
        try:
            with should.raise_exception(CustomException):
                pass
        except TestFailed as e:
            self.assertIn(" not raised", str(e))
        else:
            self.fail("should.raise_exception allowed success")

    def test_raises_exception_multiple_classes(self):
        possible_exception_types = (CustomException, OtherException)
        for x in possible_exception_types:
            with should.raise_exception(possible_exception_types):
                raise x()

        with self.assertRaises(TestFailed) as caught:
            with should.raise_exception((CustomException,)) as caught:
                pass

## boilerplate

class OtherException(BaseException):
    pass

_MESSAGE = "SOME MESSAGE HERE"


def good(*args):
    _test_assertion(args, positive=True)

def bad(*args):
    _test_assertion(args, positive=False)

def _test_assertion(args, positive):
    positive_assertion = _current_positive_assertion
    negative_assertion = _current_negative_assertion
    if not positive:
        positive_assertion, negative_assertion = negative_assertion, positive_assertion
    positive_assertion(*args)
    positive_assertion(*(args+(_MESSAGE,)))
    for msg in (None, _MESSAGE):
        try:
            negative_assertion(*(args + (msg,)))
        except TestFailed as e:
            if msg is not None:
                assert msg in str(e)
        else:
            assert False, "Assertion did not fail"


_current_positive_assertion = None
_current_negative_assertion = None

@contextmanager
def checking(assertion, negative_assertion):
    global _current_positive_assertion
    global _current_negative_assertion
    assert _current_positive_assertion is None
    _current_positive_assertion = assertion
    _current_negative_assertion = negative_assertion
    try:
        yield
    finally:
        _current_positive_assertion = None
        _current_negative_assertion = None



# pylint: disable=import-error,no-name-in-module
from __future__ import division
import collections
import itertools
import sys

from py.io import TerminalWriter
from textwrap import wrap

from .._compat import iteritems, izip
from ..conf import config
from ..log import VERBOSITIES
from ..utils.iteration import iteration
from ..utils.python import wraps
from .reporter_interface import ReporterInterface

# traceback levels
NO_TRACEBACK, SINGLE_FRAME, ALL_FRAMES, ALL_FRAMES_WITH_CONTEXT, ALL_FRAMES_WITH_CONTEXT_AND_VARS = range(
    5)


def from_verbosity(level):
    def decorator(func):
        @wraps(func)
        def new_func(self, *args, **kwargs):
            if self._verobsity_allows(level):  # pylint: disable=protected-access
                return func(self, *args, **kwargs)

        return new_func
    return decorator


class TerminalWriterWrapper(object):

    def __init__(self, file):
        super(TerminalWriterWrapper, self).__init__()
        self._writer = TerminalWriter(file=file)
        self._line = ''

    def _get_full_width(self):
        fullwidth = self._writer.fullwidth
        if sys.platform == "win32":
            # see py.io documentation for an explanation
            fullwidth -= 1

        return fullwidth

    def lsep(self, sep, msg, **kw):
        """Write a left-justified line filled with the separator until the end of the line"""

        self._do_write(
            '{0} {1}\n'.format(msg, sep * ((self._get_full_width() - 1 - len(msg)) // len(sep))), **kw)

    def write_box(self, headline, msg, **kw):
        box_width = min(self._get_full_width(), 60)
        line_width = box_width - 4

        max_headline_length = box_width - 6
        if len(headline) > max_headline_length:
            headline = headline[max_headline_length:]

        def write_line(line_to_write):
            eol_padding = box_width - (len(line_to_write) + 3)
            self._do_write('* {0}{1}*\n'.format(line_to_write, ' ' * eol_padding), **kw)

        self._do_write('\n** {0} {1}\n'.format(headline, '*' * (box_width - (len(headline) + 4))), **kw)
        for line in msg.split('\n'):
            if not line:
                write_line('')
            else:
                for sub_line in wrap(line, line_width):
                    write_line(sub_line)
        self._do_write('{0}\n\n'.format('*' * box_width), **kw)

    def sep(self, *args, **kw):
        self._line = ''
        return self._writer.sep(*args, **kw)

    def write(self, line, **kw):
        line = str(line)
        self._do_write(line, **kw)
        self._line = self._get_line_remainder(line)

    def _get_line_remainder(self, line):
        return line.rsplit('\r', 1)[-1].rsplit('\n', 1)[-1]

    def line(self, *args, **kw):
        self._writer.line(*args, **kw)
        self._line = ''

    def get_line_in_progress(self):
        return self._line

    def clear_line_in_progress(self):
        if self._line and self._writer.hasmarkup:
            self._do_write('\r')
            self._do_write(' ' * len(self._line))
            self._do_write('\r')

    def restore_line_in_progress(self):
        if self._writer.hasmarkup:
            self._do_write(self._line)

    def _do_write(self, *args, **kwargs):
        return self._writer.write(*args, **kwargs)


class ConsoleReporter(ReporterInterface):

    def __init__(self, level, stream=sys.stderr):
        super(ConsoleReporter, self).__init__()
        self._level = level
        self._stream = stream
        self._terminal = TerminalWriterWrapper(file=stream)

    def notify_before_console_output(self):
        self._terminal.clear_line_in_progress()

    def notify_after_console_output(self):
        self._terminal.restore_line_in_progress()

    def report_before_debugger(self, exc_info):
        self.notify_before_console_output()
        self._terminal.write('Exception caught in debugger: {0} {1}\n'.format(
            exc_info[0], exc_info[1]), red=True)
        self.notify_after_console_output()

    def report_collection_start(self):
        self._report_num_collected([], stillworking=True)

    def report_test_collected(self, all_tests, test):
        self._report_num_collected(all_tests, stillworking=True)

    def report_collection_end(self, collected):
        self._report_num_collected(collected, stillworking=False)

    def _report_num_collected(self, collected, stillworking):
        self._terminal.write('\r{0} tests collected{1}'.format(
            len(collected), '...' if stillworking else '   \n'), white=True, bold=True)

    def _is_verbose(self, level):
        return self._level <= level

    @from_verbosity(VERBOSITIES.ERROR)
    def report_session_start(self, session):
        self._terminal.sep('=', 'Session starts', white=True, bold=True)

    def report_session_end(self, session):

        if not self._verobsity_allows(VERBOSITIES.WARNING):
            # for concise outputs we need to break the sequence of dots...
            self._terminal.write('\n')

        header_format = self._get_session_summary_header_format(session)

        for index, (test_index, test_result, infos) in enumerate(self._iter_reported_results(session)):
            if index == 0:
                self._terminal.sep('=', 'Session Summary', **header_format)
            self._report_test_summary_header(test_index, test_result)
            self._report_additional_test_details(test_result)
            for info_reporter in infos:
                info_reporter(test_result)

        if self._verobsity_allows(VERBOSITIES.WARNING):
            self._report_result_warning_summary(session)

        msg = 'Session ended.'
        msg += ' {0} successful, {1} skipped, {2} failures, {3} errors.'.format(
            session.results.get_num_successful(
            ), session.results.get_num_skipped(),
            session.results.get_num_failures(), session.results.get_num_errors())

        not_run = session.results.get_num_not_run()
        if not_run:
            msg += ' {0} not run.'.format(not_run)

        msg += ' Total duration: {0}'.format(
            self._format_duration(session.duration))
        self._terminal.sep('=', msg, **header_format)

    def _get_session_summary_header_format(self, session):
        returned = {'bold': True}
        if session.results.is_success(allow_skips=True):
            returned.update(green=True)
        else:
            returned.update(red=True)
        return returned

    def _iter_reported_results(self, session):
        for test_index, test_result in enumerate(session.results.iter_test_results()):
            infos = self._get_result_info_generators(test_result)
            if not infos:
                continue
            yield test_index, test_result, infos

    def _report_test_summary_header(self, index, test_result):
        self._terminal.lsep(
            "=", '== #{0}: {1}'.format(index + 1, test_result.test_metadata.address))

    def _get_result_info_generators(self, test_result):
        returned = []
        if self._verobsity_allows(VERBOSITIES.ERROR) and test_result.has_errors_or_failures():
            returned.append(self._report_result_errors_failures)
        if self._verobsity_allows(VERBOSITIES.INFO) and test_result.has_skips():
            returned.append(self._report_result_skip_summary)

        return returned

    def _report_result_warning_summary(self, session):
        warnings_by_key = collections.defaultdict(list)
        for warning in session.warnings:
            warnings_by_key[warning.key].append(warning)
        for i, (warning_key, warnings) in iteration(iteritems(warnings_by_key)):
            if i.first:
                self._terminal.sep(
                    '=', 'Warnings ({0} total)'.format(len(session.warnings)), yellow=True)
            self._terminal.write(
                ' * {d[filename]}:{d[lineno]:03} -- '.format(d=warnings[0].details), yellow=True)
            self._terminal.write(
                warnings[0].details['message'], yellow=True, bold=True)
            self._terminal.write(
                ' (Repeated {0} times)\n'.format(len(warnings)), yellow=True)

    def _verobsity_allows(self, level):
        return self._level <= level

    def _report_result_errors_failures(self, test_result):
        all_errs = list(
            itertools.chain(izip(itertools.repeat("E"), test_result.get_errors()),
                            izip(itertools.repeat("F"), test_result.get_failures())))
        for index, (err_type, err) in enumerate(all_errs):
            err_header = ' - {0}/{1} {2} ({3:YYYY-MM-DD HH:mm:ss ZZ}): {4}'.format(
                index + 1,
                len(all_errs),
                err_type,
                err.time.to('local'),
                ' - {0}'.format(err.message) if not err.traceback else '')
            self._terminal.lsep(' -', err_header, red=True)
            self._report_traceback(err_type, err)

    def _report_traceback(self, err_type, err):
        traceback_level = config.root.log.traceback_level
        if not err.traceback or traceback_level == NO_TRACEBACK:
            frames = []
        elif traceback_level == SINGLE_FRAME:
            frames = [err.traceback.frames[-1]]
        else:
            frames = err.traceback.frames
        for frame_iteration, frame in iteration(frames):
            if traceback_level >= ALL_FRAMES_WITH_CONTEXT_AND_VARS:

                if not frame_iteration.first:
                    self._terminal.sep('- ')
            self._terminal.write(
                ' {0}:{1}\n'.format(frame.filename, frame.lineno), white=True, bold=True)

            if traceback_level >= ALL_FRAMES_WITH_CONTEXT_AND_VARS:
                self._write_frame_locals(frame)

            code_lines = self._write_frame_code(
                frame, include_context=(traceback_level >= ALL_FRAMES_WITH_CONTEXT))
            if frame_iteration.last:
                self._terminal.write(err_type, red=True, bold=True)
                self._terminal.write(
                    self._indent_with(err.message, 4), red=True, bold=True)
                self._terminal.write('\n')


    def _report_additional_test_details(self, result):
        if result.is_success():
            return
        detail_items = iteritems(result.get_additional_details())

        log_path = result.get_log_path()
        if log_path is not None:
            detail_items = itertools.chain(detail_items, [('Log', log_path)])

        for index, (key, value) in enumerate(detail_items):
            if index == 0:
                self._terminal.write(' - Additional Details:\n', black=True, bold=True)
            self._terminal.write('    > {0}: {1!r}\n'.format(key, value), black=True, bold=True)

    def _indent_with(self, text, indent):
        if isinstance(indent, int):
            indent = ' ' * indent
        return '\n'.join(indent + line for line in text.splitlines())

    def _report_result_skip_summary(self, result):
        msg = '\tSkipped'
        skip_reason = result.get_skips()[0]
        if skip_reason is not None:
            msg += ' ({0})'.format(skip_reason)
        msg += '\n'

        self._terminal.write(msg, yellow=True)

    def _write_frame_locals(self, frame):
        if not frame.locals and not frame.globals:
            return
        for index, (name, value) in enumerate(itertools.chain(iteritems(frame.locals), iteritems(frame.globals))):
            if index > 0:
                self._terminal.write(', ')
            self._terminal.write(
                '    {0}: '.format(name), yellow=True, bold=True)
            self._terminal.write(value['value'])
        self._terminal.write('\n\n')

    def _write_frame_code(self, frame, include_context):
        if frame.code_string:
            if include_context:
                code_lines = frame.code_string.splitlines()
            else:
                code_lines = [frame.code_line]
            line = ''
            for line_iteration, line in iteration(code_lines):
                if line_iteration.last:
                    self._terminal.write('>', white=True, bold=True)
                else:
                    self._terminal.write(' ')
                self._terminal.write(line, black=not line_iteration.last, bold=not line_iteration.last)
                self._terminal.write('\n')
            return code_lines

    @from_verbosity(VERBOSITIES.WARNING)
    def report_file_start(self, filename):
        self._file_failed = False
        self._file_has_skips = False
        if not self._verobsity_allows(VERBOSITIES.NOTICE):
            self._terminal.write(filename)
            self._terminal.write(' ')

    @from_verbosity(VERBOSITIES.WARNING)
    def report_file_end(self, filename):
        if self._verobsity_allows(VERBOSITIES.NOTICE):
            return
        self._terminal.write('  ')
        if self._file_failed:
            self._terminal.line('FAIL', red=True)
        elif self._file_has_skips:
            self._terminal.line('PASS', yellow=True)
        else:
            self._terminal.line('PASS', green=True)

    def report_test_success(self, test, result):
        if not self._verobsity_allows(VERBOSITIES.NOTICE):
            self._terminal.write('.')

    def report_test_skip_added(self, test, reason):
        self._file_has_skips = True
        if self._verobsity_allows(VERBOSITIES.NOTICE):
            self._terminal.write('Skipped: {0}\n'.format(reason), yellow=True)
        else:
            self._terminal.write('s', yellow=True)

    def report_test_error_added(self, test, error):
        self._report_test_error_failure_added(test, error, 'E')

    def report_test_failure_added(self, test, error):
        self._report_test_error_failure_added(test, error, 'F')

    def _report_test_error_failure_added(self, test, e, errtype):
        self._file_failed = True
        if not self._verobsity_allows(VERBOSITIES.NOTICE):
            self._terminal.write(errtype, red=True)
        else:
            self._terminal.write('{0}: {1}\n'.format(errtype, e), red=True)

    def report_fancy_message(self, headline, message):
        if self._verobsity_allows(VERBOSITIES.INFO):
            self._terminal.write_box(headline, message, bold=True, yellow=True)

    def report_message(self, message):
        self.notify_before_console_output()
        self._terminal.write(message)
        self._terminal.write('\n')
        self.notify_after_console_output()

    def _format_duration(self, duration):
        seconds = duration % 60
        duration /= 60
        minutes = duration % 60
        hours = duration / 60
        return '{0:02}:{1:02}:{2:02}'.format(int(hours), int(minutes), int(seconds))



from __future__ import absolute_import

import inspect
import linecache
import os
import sys
import traceback
import types


def get_traceback_string(exc_info=None):
    if exc_info is None:
        exc_info = sys.exc_info()
    exc_type, exc_value, exc_tb = exc_info  # pylint: disable=unpacking-non-sequence
    return "".join(traceback.format_exception(exc_type, exc_value, exc_tb))


def distill_traceback(tb, **kw):
    return _distill_frames(_get_tb_frames(tb), **kw)


def distill_call_stack(**kw):
    return _distill_frames(_get_sys_trace_frames(), **kw)


def _distill_frames(frames, frame_correction=0):
    returned = DistilledTraceback()
    frames = frames[:len(frames)-frame_correction+1]
    for frame in frames:
        if isinstance(frame, tuple):
            frame, lineno = frame
        else:
            lineno = None
        if _is_frame_and_below_muted(frame):
            break
        if not _is_frame_muted(frame):
            returned.frames.append(DistilledFrame(frame, lineno))
    return returned


def _get_tb_frames(tb):
    returned = []
    while tb is not None:
        returned.append((tb.tb_frame, tb.tb_lineno))
        tb = tb.tb_next
    return returned

def _get_sys_trace_frames():
    return [f[0] for f in reversed(inspect.stack()[:-1])]

def _is_frame_muted(frame):
    try:
        return _deduce_frame_function(frame) in _MUTED_LOCATIONS
    finally:
        del frame


def _deduce_frame_function(frame):
    try:
        frame_module = _deduce_frame_module(frame)
        if frame_module is None:
            return None

        return (frame_module, frame.f_code.co_name)
    finally:
        del frame


def _deduce_frame_module(frame):
    return frame.f_globals.get("__name__")


_MUTED_LOCATIONS = set([
    ("slash.core.function_test", "run"),
    ("slash.core.test", "run"),
    ("slash.exception_handling", "handling_exceptions"),
    ("slash.core.fixtures.fixture_store", "call_with_fixtures"),
    ("slash.frontend.main", "__main__"),
    ("slash.frontend.main", "main"),
    ("slash.frontend.main", "main_entry_point"),
    ("slash.frontend.slash_run", "slash_run"),
    ("slash.runner", "run_tests"),
    ("slash.runner", "_get_run_context_stack"),
    ("slash.core.cleanup_manager", "call_cleanups"),
    ("slash.core.cleanup_manager", "__call__"),
])


def _is_frame_and_below_muted(frame):
    return frame.f_globals.get("__name__") in _MUTE_BELOW_MODULES

_MUTE_BELOW_MODULES = set([
    "slash.assertions"
])


class DistilledTraceback(object):

    def __init__(self):
        super(DistilledTraceback, self).__init__()
        self.frames = []

    def to_list(self):
        return [frame.to_dict() for frame in self.frames]

    @property
    def cause(self):
        if self.frames:
            return self.frames[-1]

    def __repr__(self):
        return '\n'.join(str(frame) for frame in self.frames)


class DistilledFrame(object):

    def __init__(self, frame, lineno=None):
        super(DistilledFrame, self).__init__()
        self.filename = os.path.abspath(frame.f_code.co_filename)
        if lineno is None:
            lineno = frame.f_lineno
        self.lineno = lineno
        self.func_name = frame.f_code.co_name
        self.locals = self._capture_locals(frame)
        self.globals = self._capture_globals(frame)
        self.code_line = linecache.getline(self.filename, self.lineno).rstrip()
        self.code_string = "".join(
            linecache.getline(self.filename, lineno)
            for lineno in range(frame.f_code.co_firstlineno, self.lineno + 1)) or None

    def to_dict(self):
        serialized = {}
        for attr in ['filename', 'lineno', 'func_name', 'locals', 'globals', 'code_line', 'code_string']:
            serialized[attr] = getattr(self, attr)
        return serialized

    def _capture_globals(self, frame):
        used_globals = set(frame.f_code.co_names)
        return dict((global_name, {"value": _safe_repr(value)})
                    for global_name, value in frame.f_globals.items()
                    if global_name in used_globals and self._is_global_included(value))

    def _is_global_included(self, g):
        if isinstance(g, (types.FunctionType, types.MethodType, types.ModuleType, type)):
            return False
        return True

    def _capture_locals(self, frame):
        return dict((local_name, {"value": _safe_repr(value)})
                    for local_name, value in frame.f_locals.items()
                    if "@" not in local_name)

    def __repr__(self):
        return '{0.filename}, line {0.lineno}:\n    {0.code_line}'.format(self)


def _safe_repr(value):
    try:
        return repr(value)
    except Exception:  # pylint: disable=broad-except
        return "[Unprintable {0!r} object]".format(type(value).__name__)



import copy
import functools
import itertools
import os
import sys
import tempfile
import types
from contextlib import contextmanager

from slash._compat import StringIO
from slash.frontend.slash_run import slash_run

from ..code_formatter import CodeFormatter
from .file import File
from .slash_run_result import SlashRunResult
from .suite_strategy import BalancedStrategy
from .validation import validate_run, get_test_id_from_test_address


class Suite(object):

    def __init__(self, strategy=BalancedStrategy(), path=None, debug_info=True):
        super(Suite, self).__init__()
        self._path = path
        self.strategy = strategy
        self.debug_info = debug_info
        self.clear()

    def populate(self, num_tests=10):
        for i in range(num_tests):
            self.add_test()

    def clear(self):
        self._files = []
        self._notified = []
        self._num_method_tests = self._num_function_tests = 0
        self._slashconf = None

    @property
    def classes(self):
        return [cls for file in self._files for cls in file.classes]

    @property
    def files(self):
        return list(self._files)

    @property
    def slashconf(self):
        if self._slashconf is None:
            self._slashconf = File(self, relpath='slashconf.py')
        return self._slashconf

    def add_test(self, type=None, file=None):
        if type is None:
            type = self.strategy.get_test_type()
        if type == 'function':
            returned = self.add_function_test()
        elif type == 'method':
            returned = self.add_method_test()
        else:
            raise NotImplementedError('Unknown test type {0!r}'.format(type))  # pragma: no cover
        assert returned in self._notified
        return returned

    def add_method_test(self):
        cls = self.strategy.get_class_for_test(
            self.strategy.get_file_for_test(self))
        return cls.add_method_test()

    def add_function_test(self):
        return self.strategy.get_file_for_test(self).add_function_test()

    def notify_test_added(self, test):
        self._notified.append(test)
        if test.is_method_test():
            self._num_method_tests += 1
        else:
            self._num_function_tests += 1

    def add_file(self):
        returned = File(self)
        self._files.append(returned)
        return returned

    def get_last_file(self):
        if not self._files:
            return None
        return self._files[-1]

    def __len__(self):
        return len(self._notified)

    def __getitem__(self, idx):
        return self._notified[idx]

    def run(self, verify=True, expect_interruption=False, additional_args=(), args=None, commit=True):
        if commit:
            path = self.commit()
        report_stream = StringIO()
        returned = SlashRunResult(report_stream=report_stream)
        captured = []
        with self._capture_events(returned):
            if args is None:
                args = [path]
            args.extend(additional_args)
            try:
                returned.exit_code = slash_run(args, report_stream=report_stream,
                          app_callback=captured.append,
                          test_sort_key=self._get_test_id_from_runnable
                          )
            except (KeyboardInterrupt, SystemExit) as e:
                if isinstance(e, KeyboardInterrupt):
                    assert expect_interruption, 'KeyboardInterrupt unexpectedly raised'
                returned.exit_code = -1
                returned.error_message = str(e)
            else:
                assert not expect_interruption, 'KeyboardInterrupt did not happen'

        if captured:
            assert len(captured) == 1
            returned.session = captured[0].session

        if verify:
            validate_run(self, returned, expect_interruption)
        return returned

    def _get_test_id_from_runnable(self, test):
        return get_test_id_from_test_address(test.__slash__.address)

    @contextmanager
    def _capture_events(self, summary):

        sys.modules['__ut__'] = summary.tracker
        try:
            yield
        finally:
            sys.modules.pop('__ut__')

    def commit(self):
        path = self._path
        if path is None:
            path = tempfile.mkdtemp()
        elif not os.path.isdir(path):
            os.makedirs(path)

        files = self._files
        if self._slashconf is not None:
            files = itertools.chain(files, [self._slashconf])

        # TODO: clean up paths
        for file in files:
            with open(os.path.join(path, file.get_relative_path()), 'w') as f:
                formatter = CodeFormatter(f)
                file.write(formatter)

        return path

    # Shortcuts

    @property
    def num_method_tests(self):
        return self._num_method_tests

    @property
    def num_function_tests(self):
        return self._num_function_tests

    @property
    def method_tests(self):
        return [test for test in self._notified if test.is_method_test()]

    @property
    def function_tests(self):
        return [test for test in self._notified if not test.is_method_test()]



import functools

from .core import cleanup_manager
from .ctx import context


@functools.wraps(cleanup_manager.CleanupManager.add_cleanup)
def add_cleanup(*args, **kwargs):
    if context.session is None or context.session.cleanups is None:
        raise RuntimeError('Adding cleanups requires an active session')
    return context.session.cleanups.add_cleanup(*args, **kwargs)


def add_critical_cleanup(_func, *args, **kwargs):
    """
    Same as :func:`.add_cleanup`, only the cleanup will be called even on interrupted tests
    """
    return add_cleanup(_func, critical=True, *args, **kwargs)


def add_success_only_cleanup(_func, *args, **kwargs):
    """
    Same as :func:`.add_cleanup`, only the cleanup will be called only if the test succeeds
    """
    return add_cleanup(_func, success_only=True, *args, **kwargs)



import os

from emport import import_file
from sentinels import NOTHING

import gossip
import gossip.hooks

from .. import hooks
from .._compat import itervalues
from ..conf import config
from ..utils.marks import mark, try_get_mark
from .interface import PluginInterface


_SKIPPED_PLUGIN_METHOD_NAMES = set(dir(PluginInterface))

class IncompatiblePlugin(ValueError):
    pass

class PluginManager(object):
    def __init__(self):
        super(PluginManager, self).__init__()
        self._installed = {}
        self._active = set()
        self.install_builtin_plugins()

    def discover(self):
        """
        Iterates over all search paths and loads plugins
        """
        for search_path in config.root.plugins.search_paths:
            for path, _, filenames in os.walk(search_path):
                for filename in filenames:
                    if not filename.endswith(".py"):
                        continue
                    module = import_file(os.path.join(path, filename))
                    install_func = getattr(module, "install_plugins", None)
                    if install_func is None:
                        continue
                    install_func()

    def get_installed_plugins(self):
        """
        Returns a dict mapping plugin names to currently installed plugins
        """
        return self._installed.copy()

    def get_active_plugins(self):
        """
        Returns a dict mapping plugin names to currently active plugins
        """
        return dict(self._iterate_active_plugins())

    def _iterate_active_plugins(self):
        for active_name in self._active:
            yield (active_name, self._installed[active_name])

    def get_plugin(self, plugin_name):
        """
        Retrieves a registered plugin by name, or raises a LookupError
        """
        return self._installed[plugin_name]

    def install(self, plugin, activate=False):
        """
        Installs a plugin object to the plugin mechanism. ``plugin`` must be an object deriving from
        :class:`slash.plugins.PluginInterface`.
        """
        if not isinstance(plugin, PluginInterface):
            raise IncompatiblePlugin("Invalid plugin type: {0!r}".format(type(plugin)))
        plugin_name = plugin.get_name()
        self._installed[plugin_name] = plugin
        if activate:
            self.activate(plugin_name)

    def install_builtin_plugins(self):
        for builtin_plugin_module in self._iter_builtin_plugin_modules():
            module = __import__(
                "slash.plugins.builtin.{0}".format(builtin_plugin_module),
                fromlist=[""]
            )
            self.install(module.Plugin())

    def _iter_builtin_plugin_modules(self):
        builtin_dir = os.path.join(os.path.dirname(__file__), "builtin")
        for filename in os.listdir(builtin_dir):
            if filename.startswith("_") or filename.startswith(".") or not filename.endswith(".py"):
                continue
            yield filename[:-3]

    def uninstall(self, plugin):
        """
        Uninstalls a plugin
        """
        plugin = self._get_installed_plugin(plugin)
        try:
            self.deactivate(plugin)
        except IncompatiblePlugin:
            pass
        self._installed.pop(plugin.get_name())

    def uninstall_all(self):
        """
        Uninstalls all installed plugins
        """
        for plugin in list(itervalues(self._installed)):
            self.uninstall(plugin)

    def activate(self, plugin):
        """
        Activates a plugin, registering its hook callbacks to their respective hooks.

        :param plugin: either a plugin object or a plugin name
        """
        plugin = self._get_installed_plugin(plugin)
        plugin_name = plugin.get_name()
        self._configure(plugin)
        plugin.activate()
        for hook, callback in self._get_plugin_registrations(plugin):
            hook.register(callback, token=self._get_token(plugin_name))
        self._active.add(plugin_name)

    def deactivate(self, plugin):
        """
        Deactivates a plugin, unregistering all of its hook callbacks

        :param plugin: either a plugin object or a plugin name
        """
        plugin = self._get_installed_plugin(plugin)
        plugin_name = plugin.get_name()

        if plugin_name in self._active:
            gossip.get_global_group().unregister_token(self._get_token(plugin_name))
            self._active.discard(plugin_name)
            plugin.deactivate()
        self._unconfigure(plugin)

    def _configure(self, plugin):
        cfg = plugin.get_config()
        if cfg is not None:
            config['plugin_config'].extend({plugin.get_name(): cfg})

    def _unconfigure(self, plugin):
        plugin_config = config['plugin_config']
        if plugin.get_name() in plugin_config:
            plugin_config.pop(plugin.get_name())

    def _get_token(self, plugin_name):
        return "slash.plugins.{0}".format(plugin_name)

    def _get_installed_plugin(self, plugin):
        if isinstance(plugin, str):
            plugin_name = plugin
            plugin = self._installed[plugin_name]
        else:
            plugin_name = plugin.get_name()
        if self._installed.get(plugin_name) is not plugin:
            raise ValueError("Specified plugin is not installed!")
        return plugin

    def _get_plugin_registrations(self, plugin):
        returned = []
        unknown = []
        for method_name in dir(type(plugin)):
            if method_name in _SKIPPED_PLUGIN_METHOD_NAMES:
                continue

            method = getattr(plugin, method_name)

            if not hasattr(method, '__call__'):
                continue

            hook_name = try_get_mark(method, 'register_on', NOTHING)

            if hook_name is None:
                # asked not to register for nothing
                continue

            if hook_name is not NOTHING:
                expect_exists = False
            else:
                if method_name.startswith('_'):
                    continue
                expect_exists = True
                hook_name = "slash.{0}".format(method_name)

            try:
                if expect_exists:
                    hook = gossip.get_hook(hook_name)
                else:
                    hook = gossip.hooks.get_or_create_hook(hook_name)
                    if not hook.is_defined() and hook.group.is_strict():
                        raise LookupError()
            except LookupError:
                unknown.append(hook_name)
                continue
            assert hook is not None
            returned.append((hook, method))
        if unknown:
            raise IncompatiblePlugin("Unknown hooks: {0}".format(", ".join(unknown)))
        return returned

manager = PluginManager()

def registers_on(hook_name):
    """Marks the decorated plugin method to register on a custom hook, rather than
    the method name in the 'slash' group, which is the default behavior for plugins

    Specifying ``registers_on(None)`` means that this is not a hook entry point at all.
    """
    return mark("register_on", hook_name)

def active(plugin_class):
    """Decorator for automatically installing and activating a plugin upon definition
    """
    plugin = plugin_class()
    manager.install(plugin)
    manager.activate(plugin)

    return plugin_class



import itertools

from slash._compat import itervalues


class Cartesian(object):

    def __init__(self):
        super(Cartesian, self).__init__()
        self.sets = {}
        self._assigns = []

    def assign_all(self, source_name, target_name):
        """
        For every expected combination, assigns one key from the other
        """
        self._assigns.append((source_name, target_name))

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(attr)
        return SetMaker(self, attr)

    def __len__(self):
        if not self.sets:
            return 0

        returned = 1
        for x in itervalues(self.sets):
            returned *= len(x)
        return returned

    def check(self, iterator):
        names = list(self.sets)
        sets = [self.sets[name] for name in names]
        expected = sorted((self._build_combination(names, combination) for combination in itertools.product(*sets)), key=lambda d: sorted(d.items()))
        got = sorted(iterator, key=lambda d: sorted(d.items()))
        assert got == expected

    def _build_combination(self, names, combination):
        returned = {}
        for name, value in zip(names, combination):
            returned[name] = value
        for assign_source, assign_target in self._assigns:
            returned[assign_target] = returned[assign_source]
        return returned

class SetMaker(object):

    def __init__(self, cartesian, name):
        super(SetMaker, self).__init__()
        self.cartesian = cartesian
        self.name = name

    def make_set(self, size=3):
        assert self.name not in self.cartesian.sets
        returned = self.cartesian.sets[self.name] = ["{0}{1}".format(self.name, i) for i in range(size)]
        return returned



import colorama
import functools

class ColorStringBase(object):

    def get_colored(self):
        raise NotImplementedError() # pragma: no cover

    def __repr__(self):
        return repr(str(self))

    def __add__(self, other):
        return ColorCompoundString(self, other)

    def __radd__(self, other):
        return ColorCompoundString(other, self)

class ColorString(ColorStringBase):
    def __init__(self, string, color):
        super(ColorString, self).__init__()
        self._string = string
        self._color = color

    def __len__(self):
        return len(self._string)

    def ljust(self, *args):
        return ColorString(self._string.ljust(*args), self._color)

    @classmethod
    def get_formatter(cls, color):
        return functools.partial(cls, color=color)

    def __mod__(self, values):
        return ColorString(self._string % values, self._color)

    def __str__(self):
        return str(self._string)

    def get_colored(self):
        return "{0}{1}{2}".format(getattr(colorama.Fore, self._color.upper()), self._string, colorama.Fore.RESET) # pylint: disable=no-member

class ColorCompoundString(ColorStringBase):

    def __init__(self, *strings):
        super(ColorCompoundString, self).__init__()
        self._strings = strings

    def __str__(self):
        return ''.join(str(x) for x in  self._strings)

    def __len__(self):
        return sum(len(s) for s in self._strings)

    def ljust(self):
        raise NotImplementedError() # pragma: no cover

    def get_colored(self):
        return ''.join(s.get_colored() if isinstance(s, ColorStringBase) else s for s in self._strings)



from __future__ import print_function
from contextlib import contextmanager
import errno

from .color_string import ColorString


class Formatter(object):

    def __init__(self, stream, indentation_string=" "):
        super(Formatter, self).__init__()
        self._indentation_string = indentation_string
        self._indentation_list = []
        self._indentation = ""
        self._stream = LineTracker(stream)
        self._isatty = stream.isatty()

    def write_separator(self, length=80):
        self.writeln("-" * length)

    def writeln(self, *args, **kwargs):
        self.write(end="\n", *args, **kwargs)

    def write(self, *args, **kwargs):
        try:
            end = kwargs.pop('end', '')
            for arg in args:
                if isinstance(arg, ColorString):
                    if self._isatty:
                        arg = arg.get_colored()
                    else:
                        arg = str(arg)
                lines = str(arg).splitlines()
                for index, line in enumerate(lines):
                    if self._stream.is_line_empty():
                        self._stream.write(self._indentation)
                    self._stream.write(line)
                    if index != len(lines) - 1:
                        self._stream.write("\n")
            self._stream.write(end)
        except IOError as e:
            if e.errno not in (errno.EIO, errno.EPIPE):
                raise

    def indent(self):
        self._indent(1)

    def dedent(self):
        self._indent(-1)

    def _indent(self, increment, string=None):
        if string is None:
            string = self._indentation_string
        if increment < 0:
            del self._indentation_list[increment:]
        else:
            self._indentation_list.extend(string for x in range(increment))
        self._indentation = ''.join(self._indentation_list)

    @contextmanager
    def indented(self, increment=1, string=None):
        self._indent(increment=increment, string=string)
        try:
            yield
        finally:
            self._indent(increment=-increment)


class LineTracker(object):

    def __init__(self, stream):
        super(LineTracker, self).__init__()
        self._stream = stream
        self._empty_line = True

    def write(self, output):
        self._stream.write(output)
        if output:
            self._empty_line = output.endswith('\n')

    def is_line_empty(self):
        return self._empty_line



from slash.frontend.slash_list import slash_list
from slash._compat import StringIO

import pytest

@pytest.mark.parametrize('flag', ["--only-fixtures", "--only-tests", None])
def test_slash_list(suite, flag):
    suite.debug_info = False
    f = suite.slashconf.add_fixture()
    path = suite.commit()
    report_stream = StringIO()

    args = [path]
    if flag is not None:
        args.append(flag)
    slash_list(args, report_stream)
    assert report_stream.getvalue()





import functools
import itertools
import sys
from numbers import Number

import logbook

from .._compat import itervalues, OrderedDict
from ..ctx import context
from .. import hooks
from .error import Error
from ..exceptions import FAILURE_EXCEPTION_TYPES, SkipTest
from ..utils.exception_mark import ExceptionMarker

_logger = logbook.Logger(__name__)

_ADDED_TO_RESULT = ExceptionMarker('added_to_result')


class Result(object):

    def __init__(self, test_metadata=None):
        super(Result, self).__init__()
        self.test_metadata = test_metadata
        #: dictionary to be use by tests and plugins to store result-related information for later analysis
        self.data = {}
        self._errors = []
        self._failures = []
        self._skips = []
        self._details = {}
        self._started = False
        self._finished = False
        self._interrupted = False
        self._log_path = None

    def add_exception(self, exc_info=None):
        """Adds the currently active exception, assuming it wasn't already added to a result
        """
        if exc_info is None:
            exc_info = sys.exc_info()
        exc_class, exc_value, _ = exc_info  # pylint: disable=unpacking-non-sequence

        if _ADDED_TO_RESULT.is_exception_marked(exc_value):
            return

        _ADDED_TO_RESULT.mark_exception(exc_value)
        if isinstance(exc_value, FAILURE_EXCEPTION_TYPES):
            self.add_failure()
        elif issubclass(exc_class, Exception) and not issubclass(exc_class, SkipTest):
            self.add_error()

    def has_errors_or_failures(self):
        return bool(self._failures or self._errors)

    def get_log_path(self):
        return self._log_path

    def set_log_path(self, path):
        self._log_path = path

    def is_started(self):
        return self._started

    def is_not_run(self):
        return not self._started

    def mark_started(self):
        self._started = True

    def is_error(self):
        return bool(self._errors)

    @property
    def test_id(self):
        return self.test_metadata.id

    def is_failure(self):
        return bool(self._failures)

    def is_just_failure(self):
        """Indicates this is a pure failure, without errors involved"""
        return self.is_failure() and not self.is_error()

    def is_skip(self):
        return bool(self._skips)

    def is_success(self, allow_skips=False):
        returned = not self._errors and not self._failures and not self._interrupted
        if not allow_skips:
            returned &= not self._skips
        return returned

    def is_success_finished(self):
        return self.is_success() and self.is_finished()

    def is_finished(self):
        return self._finished

    def mark_finished(self):
        self._finished = True

    def mark_interrupted(self):
        self._interrupted = True

    def is_interrupted(self):
        return self._interrupted

    def add_error(self, e=None, frame_correction=0):
        err = self._add_error(self._errors, e, frame_correction=frame_correction + 1)
        context.reporter.report_test_error_added(context.test, err)

    def add_failure(self, e=None, frame_correction=0):
        err = self._add_error(self._failures, e, frame_correction=frame_correction + 1)
        context.reporter.report_test_failure_added(context.test, err)

    def set_test_detail(self, key, value):
        self._details[key] = value

    def _add_error(self, error_list, error=None, frame_correction=0):
        try:
            if error is None:
                error = Error.capture_exception()
                if error is None:
                    raise RuntimeError('add_error() must be called with either an argument or in an active exception')
            if not isinstance(error, Error):
                error = Error(error, frame_correction=frame_correction + 1)
            _logger.debug('Error added: {0}', error)
            error_list.append(error)
            hooks.error_added(result=self, error=error)  # pylint: disable=no-member
            return error
        except Exception:
            _logger.error("Failed to add error to result", exc_info=True)
            raise

    def add_skip(self, reason):
        self._skips.append(reason)
        context.reporter.report_test_skip_added(context.test, reason)

    def get_errors(self):
        return self._errors

    def get_failures(self):
        return self._failures

    def get_additional_details(self):
        return self._details

    def get_skips(self):
        return self._skips

    def has_skips(self):
        return bool(self._skips)

    def has_fatal_exception(self):
        return any(e.is_fatal() for e in
                   itertools.chain(self._errors, self._failures))

    def __repr__(self):
        return "< Result ({0})>".format(
            ", ".join(
                attr
                for attr in ("success", "error", "failure", "skip", "finished", "interrupted")
                if getattr(self, "is_{0}".format(attr))()
            )
        )


class GlobalResult(Result):
    pass


class SessionResults(object):

    def __init__(self, session):
        super(SessionResults, self).__init__()
        self.session = session
        self.global_result = GlobalResult()
        self._results_dict = OrderedDict()
        self._iterator = functools.partial(itervalues, self._results_dict)

    def __len__(self):
        return len(self._results_dict)

    def __repr__(self):
        return '<Results: {0} successful, {1} errors, {2} failures, {3} skips>'.format(
            self.get_num_successful(),
            self.get_num_errors(),
            self.get_num_failures(),
            self.get_num_skipped())

    def iter_all_additional_details(self):
        for result in self.iter_all_results():
            if result.get_additional_details():
                yield result, result.get_additional_details()

    def iter_all_failures(self):
        for result in self.iter_all_results():
            if result.get_failures():
                yield result, result.get_failures()

    def iter_all_errors(self):
        for result in self.iter_all_results():
            if result.get_errors():
                yield result, result.get_errors()

    @property
    def current(self):
        test_id = context.test_id
        if test_id is None:
            return self.global_result
        return self._results_dict[test_id]

    def __iter__(self):
        return self._iterator()

    def is_success(self, allow_skips=False):
        if not self.global_result.is_success():
            return False
        for result in self._iterator():
            if not result.is_finished() and not result.is_skip():
                return False
            if not result.is_success(allow_skips=allow_skips):
                return False
        return True

    def is_interrupted(self):
        return any(result.is_interrupted() for result in self._iterator())

    def get_num_results(self):
        return len(self._results_dict)

    def get_num_started(self):
        return self._count(Result.is_started, include_global=False)

    def get_num_successful(self):
        return self._count(Result.is_success_finished, include_global=False)

    def get_num_errors(self):
        return self._count(Result.is_error)

    def get_num_failures(self):
        return self._count(Result.is_just_failure)

    def get_num_skipped(self):
        return self._count(Result.is_skip)

    def get_num_not_run(self):
        return self._count(Result.is_not_run, include_global=False)

    def _count(self, pred, include_global=True):
        returned = 0
        iterator = self.iter_all_results(
        ) if include_global else self.iter_test_results()
        for result in iterator:
            if pred(result):
                returned += 1
        return returned

    def iter_test_results(self):
        return iter(self)

    def iter_all_results(self):
        return itertools.chain(self.iter_test_results(), [self.global_result])

    def create_result(self, test):
        assert test.__slash__.id not in self._results_dict
        returned = Result(test.__slash__)
        self._results_dict[test.__slash__.id] = returned
        return returned

    def get_result(self, test):
        if test.__slash__ is None:
            raise LookupError("Could not find result for {0}".format(test))
        return self._results_dict[test.__slash__.id]

    def __getitem__(self, test):
        if isinstance(test, Number):
            try:
                return next(itertools.islice(itervalues(self._results_dict), test, test + 1))
            except StopIteration:
                raise IndexError()
        return self.get_result(test)
